/**
 * This software is licensed under the Apache 2 license, quoted below.
 *
 * Copyright 2011 Julien Eluard
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     [http://www.apache.org/licenses/LICENSE-2.0]
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.github.jeluard.net;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.Proxy;
import java.net.URL;
import java.net.URLStreamHandler;
import java.net.URLStreamHandlerFactory;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

/**
 * Workarounds for {@link URLStreamHandler}.
 */
public final class URLStreamHandlers {

    private static final Logger LOGGER = Logger.getLogger(URLStreamHandlers.class.getPackage().getName());

    private static final String URL_STREAM_HANDLER_FACTORY_FIELD_NAME = "factory";
    private static final String OPEN_CONNECTION_METHOD_NAME = "openConnection";
    private static final Method OPEN_CONNECTION_METHOD;
    private static final Method OPEN_CONNECTION_PROXY_METHOD;

    static {
        try {
            OPEN_CONNECTION_METHOD = URLStreamHandler.class.getDeclaredMethod(URLStreamHandlers.OPEN_CONNECTION_METHOD_NAME, URL.class);
            OPEN_CONNECTION_METHOD.setAccessible(true);
            OPEN_CONNECTION_PROXY_METHOD = URLStreamHandler.class.getDeclaredMethod(URLStreamHandlers.OPEN_CONNECTION_METHOD_NAME, URL.class, Proxy.class);
        } catch (Exception e) {
            throw new Error("Failed to initialize", e);
        }
    }

    private URLStreamHandlers() {
    }

    /**
     * Resets currently set {@link URLStreamHandlerFactory}. Necessary to be able to set a new one.
     * @throws Exception
     */
    public static void resetURLStreamHandlerFactory() throws Exception {
        final Field field = URL.class.getDeclaredField(URLStreamHandlers.URL_STREAM_HANDLER_FACTORY_FIELD_NAME);
        field.setAccessible(true);
        field.set(null, null);
    }

    /**
     * @return current {@link URLStreamHandlerFactory}
     * @throws Exception
     */
    public static URLStreamHandlerFactory getURLStreamHandlerFactory() throws Exception {
        final Field field = URL.class.getDeclaredField(URLStreamHandlers.URL_STREAM_HANDLER_FACTORY_FIELD_NAME);
        field.setAccessible(true);
        return (URLStreamHandlerFactory) field.get(null);
    }

    /**
     * Install a new {@link URLStreamHandlerFactory} that delegates to current one.
     * This factory makes sure that created {@link URLStreamHandler} have their {@link URLStreamHandler#openConnection(URL, Proxy)} hooked so that when not implemented
     * {@link URLStreamHandler#openConnection(URL)} is called.
     *
     * @throws Exception
     */
    public static void installOpenConnectionProxy() throws Exception {
        final URLStreamHandlerFactory originalStreamFactory = URLStreamHandlers.getURLStreamHandlerFactory();

        URLStreamHandlers.resetURLStreamHandlerFactory();

        //Install our patched URLStreamHandlerFactory that will create URLStreamHandler performing 'safe' openConnection calls.
        URL.setURLStreamHandlerFactory(new URLStreamHandlerFactory() {
            @Override
            public URLStreamHandler createURLStreamHandler(final String paramString) {
                final URLStreamHandler streamHandler = originalStreamFactory.createURLStreamHandler(paramString);
                final Enhancer enhancer = new Enhancer();
                enhancer.setSuperclass(URLStreamHandler.class);
                enhancer.setCallback(new MethodInterceptor() {

                    @Override
                    public Object intercept(final Object object, final Method method, final Object[] arguments, final MethodProxy proxy) throws Throwable {
                        try {
                            //If openConnection(URL, Proxy) is called we first try to call it regularly.
                            //If this fails we fail back to openConnection(URL).
                            if (method.equals(URLStreamHandlers.OPEN_CONNECTION_PROXY_METHOD)) {
                                try {
                                    if (!method.isAccessible()) {
                                        method.setAccessible(true);
                                    }
                                    return method.invoke(streamHandler, arguments);
                                } catch (InvocationTargetException e) {
                                    final Throwable cause = e.getCause();
                                    if (!(cause instanceof UnsupportedOperationException)) {
                                        throw e;
                                    }

                                    if (URLStreamHandlers.LOGGER.isLoggable(Level.FINE)) {
                                        URLStreamHandlers.LOGGER.log(Level.FINE, "Got Exception while calling openConnection(<"+arguments[0]+">, <"+arguments[1]+">)", cause);
                                    }

                                    return URLStreamHandlers.OPEN_CONNECTION_METHOD.invoke(streamHandler, arguments[0]);
                                }
                            }
                            return proxy.invokeSuper(object, arguments);
                        } catch (InvocationTargetException e) {
                            final Throwable cause = e.getCause();
                            if (URLStreamHandlers.LOGGER.isLoggable(Level.SEVERE)) {
                                URLStreamHandlers.LOGGER.log(Level.SEVERE, "Got Exception while proxying", cause);
                            }

                            throw cause;
                        }
                    }
                });
                return (URLStreamHandler) enhancer.create();
            }
        });
    }

}